from __future__ import annotations

from abc import (ABC,
                 abstractmethod)
from reprlib import recursive_repr
from typing import (ClassVar,
                    Dict,
                    List,
                    Optional,
                    Sequence,
                    Set,
                    Tuple)

from ground.base import Relation
from ground.hints import (Point,
                          Segment)
from reprit.base import generate_repr

from .utils import (classify_overlap,
                    to_sorted_pair)


class Event(ABC):
    __slots__ = ()

    is_left: ClassVar[bool]
    left: LeftEvent
    right: RightEvent

    @property
    @abstractmethod
    def end(self) -> Point:
        """Returns end of the event."""

    @property
    @abstractmethod
    def original_end(self) -> Point:
        """Returns original end of the event."""

    @property
    @abstractmethod
    def original_start(self) -> Point:
        """Returns original start of the segment."""

    @abstractmethod
    def register_tangent(self, tangent: 'Event') -> None:
        """Registers new tangent to the event"""

    @property
    @abstractmethod
    def segments_ids(self) -> Set[int]:
        """Returns segments ids of the event."""

    @property
    @abstractmethod
    def start(self) -> Point:
        """Returns start of the event."""

    @property
    @abstractmethod
    def tangents(self) -> Sequence['Event']:
        """Returns tangents of the event."""


class LeftEvent(Event):
    @classmethod
    def from_segment(cls, segment: Segment, segment_id: int) -> LeftEvent:
        start, end = to_sorted_pair(segment.start, segment.end)
        result = LeftEvent(start, None, start, {start: {end: {segment_id}}})
        result._right = RightEvent(end, result, end)
        return result

    is_left = True

    @property
    def end(self) -> Point:
        return self.right.start

    @property
    def original_start(self) -> Point:
        return self._original_start

    @property
    def original_end(self) -> Point:
        return self.right.original_start

    @property
    def segments_ids(self) -> Set[int]:
        return self.parts_ids[self.start][self.end]

    @property
    def right(self) -> RightEvent:
        result = self._right
        assert result is not None, self
        return result

    @right.setter
    def right(self, value: RightEvent) -> None:
        self._right = value

    @property
    def start(self) -> Point:
        return self._start

    @property
    def tangents(self) -> Sequence[Event]:
        return self._tangents

    _right: Optional[RightEvent]

    __slots__ = ('parts_ids', '_original_start', '_relations_mask', '_right',
                 '_start', '_tangents')

    def __init__(self,
                 start: Point,
                 right: Optional[RightEvent],
                 original_start: Point,
                 parts_ids: Dict[Point, Dict[Point, Set[int]]]) -> None:
        self._right, self.parts_ids, self._original_start, self._start = (
            right, parts_ids, original_start, start
        )
        self._relations_mask = 0
        self._tangents = []  # type: List[Event]

    __repr__ = recursive_repr()(generate_repr(__init__))

    def divide(self, point: Point) -> Tuple[RightEvent, LeftEvent]:
        """Divides the event at given break point and returns tail."""
        segments_ids = self.segments_ids
        (self.parts_ids.setdefault(self.start, {})
         .setdefault(point, set()).update(segments_ids))
        (self.parts_ids.setdefault(point, {})
         .setdefault(self.end, set()).update(segments_ids))
        point_to_end_event = self.right.left = LeftEvent(
                point, self.right, self.original_start, self.parts_ids
        )
        point_to_start_event = self._right = RightEvent(point, self,
                                                        self.original_end)
        return point_to_start_event, point_to_end_event

    def has_only_relations(self, *relations: Relation) -> bool:
        mask = self._relations_mask
        for relation in relations:
            mask &= ~(1 << relation)
        return not mask

    def merge_with(self, other: LeftEvent) -> None:
        assert self.start == other.start and self.end == other.end
        full_relation = classify_overlap(
                other.original_start, other.original_end,
                self.original_start, self.original_end
        )
        self.register_relation(full_relation)
        other.register_relation(full_relation.complement)
        start, end = self.start, self.end
        self.parts_ids[start][end] = other.parts_ids[start][end] = (
                self.parts_ids[start][end] | other.parts_ids[start][end]
        )

    def register_tangent(self, tangent: Event) -> None:
        assert self.start == tangent.start
        self._tangents.append(tangent)

    def register_relation(self, relation: Relation) -> None:
        self._relations_mask |= 1 << relation


class RightEvent(Event):
    is_left = False

    @property
    def end(self) -> Point:
        return self.left.start

    @property
    def left(self) -> LeftEvent:
        result = self._left
        assert result is not None, self
        return result

    @left.setter
    def left(self, value: LeftEvent) -> None:
        self._left = value

    @property
    def original_end(self) -> Point:
        return self.left.original_start

    @property
    def original_start(self) -> Point:
        return self._original_start

    @property
    def segments_ids(self) -> Set[int]:
        return self.left.segments_ids

    @property
    def start(self) -> Point:
        return self._start

    @property
    def tangents(self) -> Sequence[Event]:
        return self._tangents

    _left: Optional[LeftEvent]

    __slots__ = '_left', '_original_start', '_start', '_tangents'

    def __init__(self,
                 start: Point,
                 left: Optional[LeftEvent],
                 original_start: Point) -> None:
        self._left, self._original_start, self._start = (left, original_start,
                                                         start)
        self._tangents = []  # type: List[Event]

    __repr__ = recursive_repr()(generate_repr(__init__))

    def register_tangent(self, tangent: 'Event') -> None:
        assert self.start == tangent.start
        self._tangents.append(tangent)



import typing as t

import typing_extensions as te
from ground.base import Relation
from ground.hints import Point

_HashableT = t.TypeVar('_HashableT',
                       bound=t.Hashable)


def all_unique(values: t.Iterable[_HashableT]) -> bool:
    seen: t.Set[_HashableT] = set()
    seen_add = seen.add
    for value in values:
        if value in seen:
            return False
        else:
            seen_add(value)
    return True


def classify_overlap(test_start: Point,
                     test_end: Point,
                     goal_start: Point,
                     goal_end: Point) -> Relation:
    assert test_start < test_end
    assert goal_start < goal_end
    if test_start == goal_start:
        return (Relation.COMPONENT
                if test_end < goal_end
                else (Relation.COMPOSITE
                      if goal_end < test_end
                      else Relation.EQUAL))
    elif test_end == goal_end:
        return (Relation.COMPOSITE
                if test_start < goal_start
                else Relation.COMPONENT)
    elif goal_start < test_start < goal_end:
        return (Relation.COMPONENT
                if test_end < goal_end
                else Relation.OVERLAP)
    else:
        assert test_start < goal_start < test_end
        return (Relation.COMPOSITE
                if goal_end < test_end
                else Relation.OVERLAP)


class _Comparable(te.Protocol):
    def __lt__(self, other: te.Self) -> bool:
        ...


_ComparableT = t.TypeVar('_ComparableT',
                         bound=_Comparable)


def to_sorted_pair(
        start: _ComparableT, end: _ComparableT
) -> t.Tuple[_ComparableT, _ComparableT]:
    return (start, end) if start < end else (end, start)



from setuptools import (find_packages,
                        setup)

project_base_url = 'https://github.com/lycantropos/bentley_ottmann/'


setup(packages=find_packages(exclude=('tests', 'tests.*')),
      url=project_base_url,
      download_url=project_base_url + 'archive/master.zip')



from __future__ import annotations

from functools import partial
from typing import (Any,
                    Callable,
                    Optional)

from dendroid import red_black
from dendroid.hints import KeyedSet
from ground.base import (Context,
                         Orientation)
from ground.hints import Point
from reprit.base import generate_repr

from .event import LeftEvent


class SweepLine:
    __slots__ = 'context', '_set'

    def __init__(self, context: Context) -> None:
        self.context = context
        self._set: KeyedSet[SweepLineKey, LeftEvent] = red_black.set_(
                key=partial(SweepLineKey, context.angle_orientation)
        )

    __repr__ = generate_repr(__init__)

    def add(self, event: LeftEvent) -> None:
        self._set.add(event)

    def find_equal(self, event: LeftEvent) -> Optional[LeftEvent]:
        try:
            candidate = self._set.floor(event)
        except ValueError:
            return None
        else:
            return (candidate
                    if (candidate.start == event.start
                        and candidate.end == event.end)
                    else None)

    def remove(self, event: LeftEvent) -> None:
        self._set.remove(event)

    def above(self, event: LeftEvent) -> Optional[LeftEvent]:
        try:
            return self._set.next(event)
        except ValueError:
            return None

    def below(self, event: LeftEvent) -> Optional[LeftEvent]:
        try:
            return self._set.prev(event)
        except ValueError:
            return None


class SweepLineKey:
    __slots__ = 'event', 'orienteer'

    def __init__(self,
                 orienteer: Callable[[Point, Point, Point], Orientation],
                 event: LeftEvent) -> None:
        self.event, self.orienteer = event, orienteer

    __repr__ = generate_repr(__init__)

    def __lt__(self, other: SweepLineKey) -> Any:
        """
        Checks if the segment (or at least the point) associated with event
        is lower than other's.
        """
        event, other_event = self.event, other.event
        if event is other_event:
            return False
        start, other_start = event.start, other_event.start
        end, other_end = event.end, other_event.end
        other_start_orientation = self.orienteer(start, end, other_start)
        other_end_orientation = self.orienteer(start, end, other_end)
        if other_start_orientation is other_end_orientation:
            start_x, start_y = start.x, start.y
            other_start_x, other_start_y = other_start.x, other_start.y
            if other_start_orientation is not Orientation.COLLINEAR:
                # other segment fully lies on one side
                return other_start_orientation is Orientation.COUNTERCLOCKWISE
            # segments are collinear
            elif start_y == other_start_y:
                end_x, end_y = end.x, end.y
                other_end_x, other_end_y = other_end.x, other_end.y
                if start_x != other_start_x:
                    return start_x < other_start_x
                # segments have same start
                elif end_y != other_end_y:
                    return end_y < other_end_y
                else:
                    # segments are horizontal
                    return end_x < other_end_x
            else:
                return start_y < other_start_y
        start_orientation = self.orienteer(other_start, other_end, start)
        end_orientation = self.orienteer(other_start, other_end, end)
        if start_orientation is end_orientation:
            return start_orientation is Orientation.CLOCKWISE
        elif other_start_orientation is Orientation.COLLINEAR:
            return other_end_orientation is Orientation.COUNTERCLOCKWISE
        elif start_orientation is Orientation.COLLINEAR:
            return end_orientation is Orientation.CLOCKWISE
        elif end_orientation is Orientation.COLLINEAR:
            return start_orientation is Orientation.CLOCKWISE
        else:
            return other_start_orientation is Orientation.COUNTERCLOCKWISE
